# -*- coding: utf-8 -*-
import itertools
import json
import uuid
from unittest.mock import ANY, call, patch, sentinel

import odoo
from odoo.tools import mute_logger
from odoo.tests import get_db_name, HttpCase, new_test_user
from odoo.addons.mail.tests.common import MailCommon
from odoo.addons.saas_worker.util import normalize_ip_address
from odoo.addons.openerp_enterprise.tests.common import EnterpriseDatabaseCommon, DateBefore
from odoo.addons.openerp_enterprise.models.multidb_jobqueue import MAX_JOBS_BY_BATCH


RE_ERROR_EMAIL = (r'Dear Saas User 1,(\s|<br/?>)+The invitation of users\s+that you triggered '
                  r'at \d\d/\d\d/\d\d\d\d \d\d:\d\d:\d\d has failed\.')


class TestMultidbJobQueue(EnterpriseDatabaseCommon):
    SERVER_NAMES_BY_IP = {
        '198.51.100.92': 'eu1.odoo.test',
        '203.0.113.42': 'eu2.odoo.test',
    }

    def run(self, result=None):
        # Prevent _cron_process_jobs to really commit or rollback
        # Patch configuration handlers to resolve the fake servers
        with patch.object(self.env.cr, 'commit', side_effect=self.env.flush_all), \
             patch.object(self.env.cr, 'rollback', return_value=None), \
             patch('odoo.addons.openerp_enterprise.models.multidb_jobqueue.get_server_names_by_ip',
                   return_value={normalize_ip_address(ip): host for ip, host in self.SERVER_NAMES_BY_IP.items()}):
            self.IPS_BY_SERVER_NAME = {host: normalize_ip_address(ip) for ip, host in self.SERVER_NAMES_BY_IP.items()}
            super().run(result)

    def setUp(self):
        super().setUp()

        self.OEDatabase = self.env['openerp.enterprise.database']
        self.MultidbJob = self.env['openerp.enterprise.multidb.job']
        self.MultidbJobLine = self.env['openerp.enterprise.multidb.job.line']
        self.ResUsers = self.env['res.users']

        self.user1 = new_test_user(self.env, login='saas-user1@tld.net', groups='base.group_portal',
                                   name='Saas User 1', email='user1@tld.net')

        self.database1 = self.OEDatabase.create({
            'db_uuid': uuid.uuid4(),
            'ping_ids': [odoo.fields.Command.create({
                'db_name': 'saas-database-1',
                'ip': '198.51.100.92',
                'version': '15.0+e',
                'nbr_users': 1,
                'language': 'en_US',
            })],
            'database_user_ids': [odoo.fields.Command.create({
                'user_id': self.user1.id,
                'login': 'admin@database1',
            })],
            'registered_user_ids': [odoo.fields.Command.create({
                'user_id': self.user1.id,
                'is_admin': True,
            })],
        })
        self.database2 = self.OEDatabase.create({
            'db_uuid': uuid.uuid4(),
            'ping_ids': [odoo.fields.Command.create({
                'db_name': 'saas-database-2',
                'ip': '198.51.100.92',
                'version': '15.0+e',
                'nbr_users': 1,
                'language': 'en_US',
            })],
            'database_user_ids': [odoo.fields.Command.create({
                'user_id': self.user1.id,
                'login': 'admin@database2',
            })],
            'registered_user_ids': [odoo.fields.Command.create({
                'user_id': self.user1.id,
                'is_admin': True,
            })],
        })

        self.job1 = self.MultidbJob.create([{
            'user_id': self.user1.id,
            'type': 'invite_users',
            'arguments': json.dumps({'invited_users': ['foo@bar.net', 'baz@doo.org']}),
            'line_ids': [
                odoo.fields.Command.create({'database_id': self.database1.id}),
                odoo.fields.Command.create({'database_id': self.database2.id}),
            ]
        }])

    def test_action_retry(self):
        previous_cron_triggers = self.env['ir.cron.trigger'].search([])

        for previous_state, expected_state, expected_trigger in (
            ('todo', 'todo', False),
            ('done', 'done', False),
            ('error', 'todo', True),
        ):
            with self.subTest(previous_state=previous_state):
                self.job1.line_ids.write({'state': previous_state})
                self.job1.write({'result': '{}' if previous_state == 'done' else False})

                self.job1.action_retry()

                self.assertEqual(self.job1.line_ids.mapped('state'), [expected_state] * 2)
                self.assertEqual(self.job1.state, expected_state)
                cron_triggers = self.env['ir.cron.trigger'].search([('id', 'not in', previous_cron_triggers.ids)])
                if expected_trigger:
                    self.assertRecordValues(cron_triggers, [{
                        'cron_id': self.env.ref('openerp_enterprise.cron_multidb_jobs').id,
                        'call_at': DateBefore(odoo.fields.Datetime.now()),
                    }])
                else:
                    self.assertFalse(cron_triggers)
                previous_cron_triggers |= cron_triggers

    def test_action_retry_finalize(self):
        # Special case of failed _finalize, which lets lines as 'done' but job as 'error'
        previous_cron_triggers = self.env['ir.cron.trigger'].search([])
        self.job1.line_ids.write({'state': 'done'})
        self.job1.write({'state': 'error'})

        self.job1.action_retry()

        self.assertEqual(self.job1.line_ids.mapped('state'), ['done'] * 2)
        self.assertEqual(self.job1.state, 'todo')
        cron_triggers = self.env['ir.cron.trigger'].search([('id', 'not in', previous_cron_triggers.ids)])
        self.assertRecordValues(cron_triggers, [{
            'cron_id': self.env.ref('openerp_enterprise.cron_multidb_jobs').id,
            'call_at': DateBefore(odoo.fields.Datetime.now()),
        }])


class TestMultidbJobQueueCronProcessJobs(TestMultidbJobQueue, MailCommon):

    def run(self, result=None):
        module_name = 'odoo.addons.openerp_enterprise.models.multidb_jobqueue'
        # Force the USER constant to 'odoo' for get_port to behave properly
        # Add mocks on called methods to check calls
        with patch('odoo.addons.saas_worker.metabase.USER', new='odoo'), \
             patch(f'{module_name}.OpenerpEnterpriseMultidbJobLine._run_invite_users', autospec=True) as mock_run_invite_users, \
             patch(f'{module_name}.OpenerpEnterpriseMultidbJobLine._finalize_invite_users', autospec=True) as mock_finalize:
            self.mock_run_invite_users = mock_run_invite_users
            self.mock_finalize_invite_users = mock_finalize
            return super().run(result)

    def setUp(self):
        super().setUp()
        # By default, let's pretend all calls go well
        self._run_invite_users_iter = itertools.cycle([('done', 'true')])
        self._finalize_invite_users_iter = itertools.cycle([('done', 'true')])

        def _run_invite_users(line_ids, hostname, port):
            state, result = next(self._run_invite_users_iter)
            line_ids.write({
                'result': result,
                'state': state,
            })

        def _finalize_invite_users(line_ids):
            state, result = next(self._finalize_invite_users_iter)
            line_ids.job_id.write({
                'result': result,
                'state': state,
            })

        self.mock_run_invite_users.side_effect = _run_invite_users
        self.mock_finalize_invite_users.side_effect = _finalize_invite_users

    def test_run_jobs(self):
        """All-went-good scenario"""
        with self.mock_mail_gateway():
            self.MultidbJob._cron_process_jobs()

        self.mock_run_invite_users.assert_called_once_with(self.job1.line_ids, self.IPS_BY_SERVER_NAME['eu1.odoo.test'], 42150)
        self.mock_finalize_invite_users.assert_called_once_with(self.job1.line_ids)

        self.assertEqual(self.job1.state, 'done')
        self.assertEqual(self.job1.line_ids.mapped('state'), ['done', 'done'])

        self.assertNotSentEmail()

    def test_run_jobs_error_run(self):
        """What happens when the run method fails"""
        self._run_invite_users_iter = itertools.cycle([('error', 'false')])

        with self.mock_mail_gateway():
            self.MultidbJob._cron_process_jobs()

        self.mock_run_invite_users.assert_called_once_with(self.job1.line_ids, self.IPS_BY_SERVER_NAME['eu1.odoo.test'], 42150)
        self.mock_finalize_invite_users.assert_not_called()

        # the job and the failing database jobs are set to state 'error'
        self.assertEqual(self.job1.state, 'error')
        self.assertEqual(self.job1.line_ids.mapped('state'), ['error', 'error'])

        # an error email is sent to the initiator user
        found_mail = self._find_mail_mail_wemail('"Saas User 1" <user1@tld.net>', 'outgoing', author=self.env.user.partner_id)
        self.assertRegex(found_mail.body_html, RE_ERROR_EMAIL)

    @mute_logger('odoo.addons.openerp_enterprise.models.multidb_jobqueue')
    def test_run_jobs_exception_run(self):
        """What happens when the run method raises an exception"""
        self.mock_run_invite_users.side_effect = Exception('Something bad happened')

        with self.mock_mail_gateway():
            self.MultidbJob._cron_process_jobs()

        self.mock_run_invite_users.assert_called_once_with(self.job1.line_ids, self.IPS_BY_SERVER_NAME['eu1.odoo.test'], 42150)
        self.mock_finalize_invite_users.assert_not_called()

        # the job and the failing database jobs are set to state 'error'
        self.assertEqual(self.job1.state, 'error')
        self.assertEqual(self.job1.line_ids.mapped('state'), ['error', 'error'])
        self.assertIn('Exception: Something bad happened\n', self.job1.line_ids[0].result)
        self.assertIn('Exception: Something bad happened\n', self.job1.line_ids[1].result)

        # an error email is sent to the initiator user
        found_mail = self._find_mail_mail_wemail('"Saas User 1" <user1@tld.net>', 'outgoing', author=self.env.user.partner_id)
        self.assertRegex(found_mail.body_html, RE_ERROR_EMAIL)

    def test_run_jobs_finalize_when_all_lines_done(self):
        self.job1.line_ids.write({'state': 'done'})
        self.assertEqual(self.job1.state, 'todo')

        with self.mock_mail_gateway():
            self.MultidbJob._cron_process_jobs()

        self.mock_run_invite_users.assert_not_called()
        self.mock_finalize_invite_users.assert_called_once_with(self.job1.line_ids)

        self.assertEqual(self.job1.state, 'done')

    def test_run_jobs_error_finalize(self):
        """What happens when the finalize method fails"""
        self._finalize_invite_users_iter = iter([('error', 'false')])

        with self.mock_mail_gateway():
            self.MultidbJob._cron_process_jobs()

        self.mock_run_invite_users.assert_called_once_with(self.job1.line_ids, self.IPS_BY_SERVER_NAME['eu1.odoo.test'], 42150)
        self.mock_finalize_invite_users.assert_called_once_with(self.job1.line_ids)

        self.assertEqual(self.job1.state, 'error')
        self.assertEqual(self.job1.line_ids.mapped('state'), ['done', 'done'])

        # recomputing the state should leave the job in error
        self.job1._compute_state()
        self.assertEqual(self.job1.state, 'error')

        # an error email is sent to the initiator user
        found_mail = self._find_mail_mail_wemail('"Saas User 1" <user1@tld.net>', 'outgoing', author=self.env.user.partner_id)
        self.assertRegex(found_mail.body_html, r'Your Action on multiple databases</span>(\s|<br/?>)*'
                                               r'<span[^>]*>\s*Invitation of users')
        self.assertRegex(found_mail.body_html, RE_ERROR_EMAIL)
        self.assertRegex(found_mail.body_html, r'Please retry later, or contact <a[^>]*>Odoo Support</a> if the error persists\.')

    @mute_logger('odoo.addons.openerp_enterprise.models.multidb_jobqueue')
    def test_run_jobs_exception_finalize(self):
        """What happens when the finalize method raises an exception"""
        self.mock_finalize_invite_users.side_effect = Exception('Something went wrong')

        with self.mock_mail_gateway():
            self.MultidbJob._cron_process_jobs()

        self.mock_run_invite_users.assert_called_once_with(self.job1.line_ids, self.IPS_BY_SERVER_NAME['eu1.odoo.test'], 42150)
        self.mock_finalize_invite_users.assert_called_once_with(self.job1.line_ids)

        self.assertEqual(self.job1.state, 'error')
        self.assertEqual(self.job1.line_ids.mapped('state'), ['done', 'done'])

        # an error email is sent to the initiator user
        found_mail = self._find_mail_mail_wemail('"Saas User 1" <user1@tld.net>', 'outgoing', author=self.env.user.partner_id)
        self.assertRegex(found_mail.body_html, RE_ERROR_EMAIL)

    def test_run_jobs_multiserver(self):
        """What happens when the databases are hosted on different servers"""
        self.database2.ping_ids.ip = '203.0.113.42'

        with self.mock_mail_gateway():
            self.MultidbJob._cron_process_jobs()

        self.mock_run_invite_users.assert_has_calls([
            call(self.job1.line_ids[0], self.IPS_BY_SERVER_NAME['eu1.odoo.test'], 42150),
            call(self.job1.line_ids[1], self.IPS_BY_SERVER_NAME['eu2.odoo.test'], 42150),
        ])
        self.mock_finalize_invite_users.assert_called_once_with(self.job1.line_ids)

        self.assertEqual(self.job1.state, 'done')
        self.assertEqual(self.job1.line_ids.mapped('state'), ['done', 'done'])

        self.assertNotSentEmail()

    def test_run_jobs_multiserver_above_max_jobs_by_batch(self):
        """Check that we don't process more than MAX_JOBS_BY_BATCH per cron run."""
        previous_cron_triggers = self.env['ir.cron.trigger'].search([])
        self.database2.ping_ids.ip = '203.0.113.42'

        # Special case where there is more than MAX_JOBS_BY_BATCH databases on multiple servers
        # in the same multidb job
        self.job1.write({
            'line_ids': (
                [odoo.fields.Command.clear()]
                + [odoo.fields.Command.create({'database_id': self.database1.id})] * (MAX_JOBS_BY_BATCH + 1)
                + [odoo.fields.Command.create({'database_id': self.database2.id})] * 5
            )
        })

        with self.mock_mail_gateway():
            self.MultidbJob._cron_process_jobs()

        # Check that we did call the run method only once with the maximum lines limit
        self.assertEqual(self.mock_run_invite_users.call_count, 1)
        self.assertEqual(len(self.mock_run_invite_users.call_args[0][0]), MAX_JOBS_BY_BATCH)
        # Check that job is still 'todo' and only lines up to MAX_JOBS_BY_BATCH are done
        self.assertEqual(self.job1.state, 'todo')
        self.assertEqual(
            self.job1.line_ids.mapped('state'),
            (
                # should have processed up to MAX_JOBS_BY_BATCH
                ['done'] * MAX_JOBS_BY_BATCH
                # the remaining should stay as todo (for next run)
                + ['todo'] * (1 + 5)
            ),
            "We should only have processed up to MAX_JOBS_BY_BATCH job lines",
        )
        # Ensure cron did re-trigger itself
        cron_triggers = self.env['ir.cron.trigger'].search([('id', 'not in', previous_cron_triggers.ids)])
        self.assertRecordValues(cron_triggers, [{
            'cron_id': self.env.ref('openerp_enterprise.cron_multidb_jobs').id,
            'call_at': DateBefore(odoo.fields.Datetime.now()),
        }])

    def test_run_jobs_multiversion(self):
        """What happens when the databases have different Odoo versions"""
        self.database2.ping_ids.version = '10.saas~14+e'

        with self.mock_mail_gateway():
            self.MultidbJob._cron_process_jobs()

        self.mock_run_invite_users.assert_has_calls([
            call(self.job1.line_ids[0], self.IPS_BY_SERVER_NAME['eu1.odoo.test'], 42150),
            call(self.job1.line_ids[1], self.IPS_BY_SERVER_NAME['eu1.odoo.test'], 9469),
        ])
        self.mock_finalize_invite_users.assert_called_once_with(self.job1.line_ids)

        self.assertEqual(self.job1.state, 'done')
        self.assertEqual(self.job1.line_ids.mapped('state'), ['done', 'done'])

        self.assertNotSentEmail()

    def test_run_jobs_multiserver_error_run(self):
        """What happens when the databases are hosted on different servers, with one giving an error"""
        self.database2.ping_ids.ip = '203.0.113.42'
        self._run_invite_users_iter = iter([('error', 'false'), ('done', 'true')])

        with self.mock_mail_gateway():
            self.MultidbJob._cron_process_jobs()

        self.mock_run_invite_users.assert_has_calls([
            call(self.job1.line_ids[0], self.IPS_BY_SERVER_NAME['eu1.odoo.test'], 42150),
            call(self.job1.line_ids[1], self.IPS_BY_SERVER_NAME['eu2.odoo.test'], 42150),
        ])
        self.mock_finalize_invite_users.assert_not_called()

        # Only one database job is marked as error, as well as the job
        self.assertEqual(self.job1.state, 'error')
        self.assertEqual(self.job1.line_ids.mapped('state'), ['error', 'done'])
        found_mail = self._find_mail_mail_wemail('"Saas User 1" <user1@tld.net>', 'outgoing', author=self.env.user.partner_id)
        self.assertRegex(found_mail.body_html, RE_ERROR_EMAIL)

        # A second call won't run the job again
        self.mock_run_invite_users.reset_mock()
        self.mock_finalize_invite_users.reset_mock()

        with self.mock_mail_gateway():
            self.MultidbJob._cron_process_jobs()

        self.mock_run_invite_users.assert_not_called()
        self.mock_finalize_invite_users.assert_not_called()
        self.assertNotSentEmail()

    def test_run_jobs_error_continue_to_next(self):
        """When a job ends with an error, the execution continues to the next job"""
        job2 = self.job1.copy({'line_ids': [odoo.fields.Command.link(line.copy().id) for line in self.job1.line_ids]})
        job3 = self.job1.copy({'line_ids': [odoo.fields.Command.link(line.copy().id) for line in self.job1.line_ids]})
        self._run_invite_users_iter = iter([('error', 'false'), ('done', 'true'), ('done', 'true')])
        self._finalize_invite_users_iter = iter([('error', 'false'), ('done', 'true')])

        with self.mock_mail_gateway():
            self.MultidbJob._cron_process_jobs()

        self.assertEqual(self.job1.state, 'error')
        self.assertEqual(self.job1.line_ids.mapped('state'), ['error', 'error'])

        self.assertEqual(job2.state, 'error')
        self.assertEqual(job2.line_ids.mapped('state'), ['done', 'done'])

        self.assertEqual(job3.state, 'done')
        self.assertEqual(job3.line_ids.mapped('state'), ['done', 'done'])

    def test_run_jobs_db_not_found(self):
        """When a database is not found, it is ignored and _finalize is called"""
        self.database2.ping_ids.ip = '203.0.113.42'
        self._run_invite_users_iter = iter([('db_not_found', 'false'), ('done', 'true')])

        with self.mock_mail_gateway():
            self.MultidbJob._cron_process_jobs()

        self.mock_run_invite_users.assert_has_calls([
            call(self.job1.line_ids[0], self.IPS_BY_SERVER_NAME['eu1.odoo.test'], 42150),
            call(self.job1.line_ids[1], self.IPS_BY_SERVER_NAME['eu2.odoo.test'], 42150),
        ])
        self.mock_finalize_invite_users.assert_called_once_with(self.job1.line_ids)

        self.assertEqual(self.job1.state, 'done')
        self.assertEqual(self.job1.line_ids.mapped('state'), ['db_not_found', 'done'])

        self.assertNotSentEmail()

    @patch('odoo.addons.openerp_enterprise.models.multidb_jobqueue.MAX_JOBS_BY_BATCH', new=10)
    def test_run_jobs_in_small_batches(self):
        """
        Jobs should be executed by relatively small batches in order to avoid timeouts.
        Another cron run should be immediately triggered if more work is needed.
        """
        for i in range(3, 15 + 1):
            database = self.OEDatabase.create({
                'db_uuid': uuid.uuid4(),
                'ping_ids': [odoo.fields.Command.create({
                    'db_name': f'saas-database-{i}',
                    'ip': '198.51.100.92',
                    'version': '15.0',
                    'nbr_users': 1,
                    'language': 'en_US',
                })],
                'database_user_ids': [odoo.fields.Command.create({
                    'user_id': self.user1.id,
                    'login': f'admin@database{i}',
                })],
            })
            self.job1.write({
                'line_ids': [
                    odoo.fields.Command.create({'database_id': database.id}),
                ],
            })
        previous_cron_triggers = self.env['ir.cron.trigger'].search([])

        # First run should process only 10 jobs
        with self.mock_mail_gateway():
            self.MultidbJob._cron_process_jobs()

        self.assertEqual(self.job1.state, 'todo')
        self.assertEqual(self.job1.line_ids.mapped('state'), ['done'] * 10 + ['todo'] * 5)

        self.mock_run_invite_users.assert_called_once_with(self.job1.line_ids[:10], self.IPS_BY_SERVER_NAME['eu1.odoo.test'], 42150)
        self.mock_finalize_invite_users.assert_not_called()

        cron_triggers = self.env['ir.cron.trigger'].search([('id', 'not in', previous_cron_triggers.ids)])
        self.assertTrue(cron_triggers)
        self.assertRecordValues(cron_triggers, [{
            'cron_id': self.env.ref('openerp_enterprise.cron_multidb_jobs').id,
            'call_at': DateBefore(odoo.fields.Datetime.now()),
        }])
        previous_cron_triggers |= cron_triggers

        # Second run should process the 5 other jobs
        self.mock_run_invite_users.reset_mock()

        with self.mock_mail_gateway():
            self.MultidbJob._cron_process_jobs()

        self.assertEqual(self.job1.state, 'done')
        self.assertEqual(self.job1.line_ids.mapped('state'), ['done'] * 15)

        self.mock_run_invite_users.assert_called_once_with(self.job1.line_ids[10:], self.IPS_BY_SERVER_NAME['eu1.odoo.test'], 42150)
        self.mock_finalize_invite_users.assert_called_once_with(self.job1.line_ids)

        cron_triggers = self.env['ir.cron.trigger'].search([('id', 'not in', previous_cron_triggers.ids)])
        self.assertFalse(cron_triggers)


class TestMultidbJobQueueInviteUsers(TestMultidbJobQueue, HttpCase, MailCommon):

    def setUp(self):
        super().setUp()
        self.job1.write({
            'type': 'invite_users',
            'arguments': json.dumps({
                'invited_users': ['foo@bar.net', 'baz@doo.org'],
            }),
        })

    def test_controller(self):
        self.authenticate(self.user1.login, self.user1.login)

        previous_jobs = self.MultidbJob.search([])
        previous_cron_triggers = self.env['ir.cron.trigger'].search([])

        data = {'params': {
            'db_uuids': [self.database1.db_uuid, self.database2.db_uuid],
            'invitees': ['new_user@domain.tld'],
        }}
        headers = {
            'Content-Type': 'application/json',
        }
        res = self.url_open('/odoo-enterprise/invite-users-multidb', data=json.dumps(data), headers=headers)
        res.raise_for_status()
        self.assertEqual(res.json()['result'], {'result': True})

        # A new user has been created
        new_user = self.ResUsers.search([('login', '=', 'new_user@domain.tld')])
        self.assertRecordValues(new_user, [{
            'login': 'new_user@domain.tld',
            'name': 'new_user@domain.tld',
        }])

        # A new job has been created to invite the user on the databases
        new_jobs = self.MultidbJob.search([]) - previous_jobs
        self.assertRecordValues(new_jobs, [{
            'user_id': self.user1.id,
            'impostor_id': False,
            'type': 'invite_users',
            'arguments': json.dumps({
                'invited_users': ['new_user@domain.tld'],
            }),
            'state': 'todo',
        }])
        self.assertRecordValues(new_jobs.line_ids, [{
            'job_id': new_jobs.id,
            'database_id': self.database1.id,
            'state': 'todo',
        }, {
            'job_id': new_jobs.id,
            'database_id': self.database2.id,
            'state': 'todo',
        }])

        # The cron job has been triggered
        cron_triggers = self.env['ir.cron.trigger'].search([]) - previous_cron_triggers
        self.assertRecordValues(cron_triggers, [{
            'cron_id': self.env.ref('openerp_enterprise.cron_multidb_jobs').id,
            'call_at': DateBefore(odoo.fields.Datetime.now()),
        }])

    def test_controller_with_impostor(self):
        some_admin = new_test_user(self.env, login='some_admin', groups='base.group_system')
        self.authenticate(self.user1.login, self.user1.login)

        # Patch the session with impostor info
        self.session.impostor = some_admin.id
        self.session.impostor_info = some_admin.get_support_info()
        odoo.http.root.session_store.save(self.session)

        previous_jobs = self.MultidbJob.search([])

        data = {'params': {
            'db_uuids': [self.database1.db_uuid, self.database2.db_uuid],
            'invitees': ['new_user@domain.tld'],
        }}
        headers = {
            'Content-Type': 'application/json',
        }
        res = self.url_open('/odoo-enterprise/invite-users-multidb', data=json.dumps(data), headers=headers)
        res.raise_for_status()
        self.assertEqual(res.json()['result'], {'result': True})

        new_jobs = self.MultidbJob.search([]) - previous_jobs
        self.assertRecordValues(new_jobs, [{
            'user_id': self.user1.id,
            'impostor_id': some_admin.id,
        }])

    def test_controller_non_admin(self):
        self.database2.registered_user_ids.filtered(lambda x: x.user_id == self.user1).is_admin = False

        self.authenticate(self.user1.login, self.user1.login)

        data = {'params': {
            'db_uuids': [self.database1.db_uuid, self.database2.db_uuid],
            'invitees': ['new_user@domain.tld'],
        }}
        headers = {
            'Content-Type': 'application/json',
        }
        res = self.url_open('/odoo-enterprise/invite-users-multidb', data=json.dumps(data), headers=headers)
        res.raise_for_status()
        self.assertEqual(res.json()['result'], {'result': False, 'error_code': 'non_admin', 'error_data': ['saas-database-2']})

    @patch('requests.post')
    @patch('odoo.addons.openerp_enterprise.models.multidb_jobqueue.SaasAuth')
    @patch('odoo.addons.openerp_enterprise.models.auth_oauth2_token.AuthOauth2Token._get_trial_token')
    def test_run_invite_users(self, mock_get_trial_token, MockSaasAuth, mock_requests_post):
        mock_requests_post.return_value.json.return_value = {'result': {
            self.database1.db_uuid: {'result': True},
            self.database2.db_uuid: {'result': True},
        }}
        token1, token2 = sentinel.token1, sentinel.token2
        mock_get_trial_token.side_effect = [token1, token2]

        # invitees are created by the controller, before the job is created
        invitee1 = new_test_user(self.env, login='foo@bar.net', groups='base.group_portal')
        invitee2 = new_test_user(self.env, login='baz@doo.org', groups='base.group_portal')

        with self.mock_mail_gateway():
            self.job1.line_ids._run_invite_users(self.IPS_BY_SERVER_NAME['eu1.odoo.test'], 1337)

        session_info = {
            'impostor': False,
            'login': 'saas-user1@tld.net',
            'uid': self.user1.id,
        }
        mock_get_trial_token.assert_has_calls([
            call(self.user1.id, self.database1.db_uuid, session_info=session_info),
            call(self.user1.id, self.database2.db_uuid, session_info=session_info),
        ])

        expected_data = {'params': {
            'databases': [
                {'db_uuid': self.database1.db_uuid, 'oauth_token': token1, 'user_login': 'admin@database1'},
                {'db_uuid': self.database2.db_uuid, 'oauth_token': token2, 'user_login': 'admin@database2'},
            ],
            'invitees': [
                {'mail': invitee1.login, 'name': invitee1.name, 'oauth_uid': invitee1.id},
                {'mail': invitee2.login, 'name': invitee2.name, 'oauth_uid': invitee2.id},
            ],
        }}
        mock_requests_post.assert_called_once_with('http://eu1.odoo.test/saas_worker/multi/create_invitation/json',
                                                   headers={'X-Odoo-Port': '1337'}, json=expected_data,
                                                   auth=MockSaasAuth(), timeout=30)
        self.assertNotSentEmail()
        self.assertEqual(self.job1.line_ids.mapped('state'), ['done', 'done'])
        self.assertEqual(self.job1.line_ids.mapped('result'), ['true', 'true'])

    @patch('requests.post')
    @patch('odoo.addons.openerp_enterprise.models.multidb_jobqueue.SaasAuth')
    @patch('odoo.addons.openerp_enterprise.models.auth_oauth2_token.AuthOauth2Token._get_trial_token')
    def test_run_invite_users_fails_db(self, mock_get_trial_token, MockSaasAuth, mock_requests_post):
        mock_requests_post.return_value.json.return_value = {'result': {
            self.database1.db_uuid: {'result': False, 'some': 'explanation'},
            self.database2.db_uuid: {'result': True},
        }}

        with self.mock_mail_gateway():
            self.job1.line_ids._run_invite_users(self.IPS_BY_SERVER_NAME['eu1.odoo.test'], 1337)

        self.assertEqual(self.job1.line_ids.mapped('state'), ['error', 'done'])
        self.assertEqual(self.job1.line_ids.mapped('result'), [
            '{"result": false, "some": "explanation"}',
            'true',
        ])

    @patch('requests.post')
    @patch('odoo.addons.openerp_enterprise.models.multidb_jobqueue.SaasAuth')
    @patch('odoo.addons.openerp_enterprise.models.auth_oauth2_token.AuthOauth2Token._get_trial_token')
    def test_run_invite_users_fails_db_not_found(self, mock_get_trial_token, MockSaasAuth, mock_requests_post):
        mock_requests_post.return_value.json.return_value = {'result': {
            self.database1.db_uuid: {'result': False, 'error_code': 'db_not_found'},
            self.database2.db_uuid: {'result': True},
        }}

        with self.mock_mail_gateway():
            self.job1.line_ids._run_invite_users(self.IPS_BY_SERVER_NAME['eu1.odoo.test'], 1337)

        self.assertEqual(self.job1.line_ids.mapped('state'), ['db_not_found', 'done'])
        self.assertEqual(self.job1.line_ids.mapped('result'), [
            '{"result": false, "error_code": "db_not_found"}',
            'true',
        ])

    @patch('requests.post')
    @patch('odoo.addons.openerp_enterprise.models.multidb_jobqueue.SaasAuth')
    @patch('odoo.addons.openerp_enterprise.models.auth_oauth2_token.AuthOauth2Token._get_trial_token')
    def test_run_invite_users_general_failure(self, mock_get_trial_token, MockSaasAuth, mock_requests_post):
        mock_requests_post.return_value.json.return_value = {'result': False, 'any': 'explanation'}

        with self.mock_mail_gateway():
            self.job1.line_ids._run_invite_users(self.IPS_BY_SERVER_NAME['eu1.odoo.test'], 1337)

        self.assertEqual(self.job1.line_ids.mapped('state'), ['error', 'error'])
        self.assertEqual(self.job1.line_ids.mapped('result'), [json.dumps({'result': False, 'any': 'explanation'})] * 2)

    def test_finalize_invite_users(self):
        self.job1.line_ids.write({
            'result': 'true',
            'state': 'done',
        })
        self.assertFalse(self.job1.result)
        self.assertEqual(self.job1.state, 'todo')

        new_test_user(self.env, name='foo@bar.net', email='foo@bar.net', login='foo@bar.net', groups='base.group_portal')
        new_test_user(self.env, name='baz@doo.org', email='baz@doo.org', login='baz@doo.org', groups='base.group_portal',
                      log_ids=[odoo.fields.Command.create({})])  # baz has already activated its account

        with self.mock_mail_gateway():
            self.job1.line_ids._finalize_invite_users()

        self.assertTrue(self.job1.result)
        self.assertEqual(self.job1.state, 'done')

        # Check that an invitation email is sent to each of the invitees
        invitee_mail_content = 'You have been invited by Saas User 1 to connect on several Odoo databases:'

        # - newly created users receive a link to set up their password
        found_mail = self._find_mail_mail_wemail('"foo@bar.net" <foo@bar.net>', 'sent', author=self.env.user.partner_id)
        self.assertIn(invitee_mail_content, found_mail.body_html)
        self.assertIn('Set up your password', found_mail.body_html)

        # - other users receive a link to their database dashboard
        found_mail = self._find_mail_mail_wemail('"baz@doo.org" <baz@doo.org>', 'sent', author=self.env.user.partner_id)
        self.assertIn(invitee_mail_content, found_mail.body_html)
        self.assertIn('See All Your Databases', found_mail.body_html)

        # Check that a confirmation email is sent to the initiator user
        confirmation_mail_content = 'Your invitations to several databases have been processed. These users:'
        self.assertMailMailWEmails(['"Saas User 1" <user1@tld.net>'], 'outgoing', author=self.env.user.partner_id,
                                   content=confirmation_mail_content)

    @patch('requests.post')
    @patch('odoo.addons.openerp_enterprise.models.multidb_jobqueue.SaasAuth')
    @patch('odoo.addons.openerp_enterprise.models.auth_oauth2_token.AuthOauth2Token._create_token')
    def test_run_invite_users_impostor(self, mock_create_token, MockSaasAuth, mock_requests_post):
        some_admin = new_test_user(self.env, login='some_admin', groups='base.group_system')
        mock_create_token.return_value = {'access_token': 'some_token'}

        mock_requests_post.return_value.json.return_value = {'result': {
            self.database1.db_uuid: {'result': True},
            self.database2.db_uuid: {'result': True},
        }}
        self.job1.write({
            'impostor_id': some_admin.id,
        })

        self.job1.line_ids._run_invite_users(self.IPS_BY_SERVER_NAME['eu1.odoo.test'], 1337)

        session_info = {
            'impostor': some_admin.id,
            'login': 'saas-user1@tld.net',
            'uid': self.user1.id,
        }
        expected_args = call(ANY, self.user1.id, 'trial', False, False, validity_seconds=ANY, session_info=session_info)
        mock_create_token.assert_has_calls([expected_args, expected_args])

    @patch('requests.post')
    @patch('odoo.addons.openerp_enterprise.models.multidb_jobqueue.SaasAuth')
    @patch('odoo.addons.openerp_enterprise.models.auth_oauth2_token.AuthOauth2Token._get_trial_token')
    def test_run_invite_users_db_not_found(self, mock_get_trial_token, MockSaasAuth, mock_requests_post):
        """When a database is not found, it is ignored and _finalize is called"""
        mock_requests_post.return_value.json.return_value = {'result': {
            self.database1.db_uuid: {'result': False, 'error_code': 'db_not_found'},
            self.database2.db_uuid: {'result': True},
        }}

        with self.mock_mail_gateway():
            self.job1.line_ids._run_invite_users(self.IPS_BY_SERVER_NAME['eu1.odoo.test'], 1337)

        self.assertEqual(self.job1.state, 'todo')
        self.assertEqual(self.job1.line_ids.mapped('state'), ['db_not_found', 'done'])
        self.assertNotSentEmail()

        with self.mock_mail_gateway():
            self.job1.line_ids._finalize_invite_users()

        self.assertEqual(self.job1.state, 'done')
        self.assertEqual(self.job1.line_ids.mapped('state'), ['db_not_found', 'done'])

        found_mail = self._find_mail_mail_wemail('"Saas User 1" <user1@tld.net>', 'outgoing', author=self.env.user.partner_id)
        self.assertIn('Your invitations to several databases have been processed. These users:', found_mail.body_html)
        self.assertIn('However, the action could not be processed on these databases:', found_mail.body_html)
        self.assertIn('saas-database-1: database not found', found_mail.body_html)

    @patch('requests.post')
    @patch('odoo.addons.openerp_enterprise.models.multidb_jobqueue.SaasAuth')
    @patch('odoo.addons.openerp_enterprise.models.auth_oauth2_token.AuthOauth2Token._get_trial_token')
    def test_run_invite_users_all_dbs_not_found(self, mock_get_trial_token, MockSaasAuth, mock_requests_post):
        """When all databases are not found, _finalize should still send an email to the initiating user"""
        mock_requests_post.return_value.json.return_value = {'result': {
            self.database1.db_uuid: {'result': False, 'error_code': 'db_not_found'},
            self.database2.db_uuid: {'result': False, 'error_code': 'db_not_found'},
        }}

        with self.mock_mail_gateway():
            self.job1.line_ids._run_invite_users(self.IPS_BY_SERVER_NAME['eu1.odoo.test'], 1337)

        self.assertEqual(self.job1.state, 'todo')
        self.assertEqual(self.job1.line_ids.mapped('state'), ['db_not_found', 'db_not_found'])
        self.assertNotSentEmail()

        with self.mock_mail_gateway():
            self.job1.line_ids._finalize_invite_users()

        self.assertTrue(self.job1.result)
        self.assertEqual(self.job1.state, 'done')
        self.assertEqual(self.job1.line_ids.mapped('state'), ['db_not_found', 'db_not_found'])

        found_mail = self._find_mail_mail_wemail('"Saas User 1" <user1@tld.net>', 'outgoing', author=self.env.user.partner_id)
        self.assertNotIn('Your invitations to several databases have been processed. These users:', found_mail.body_html)
        self.assertIn('However, the action could not be processed on these databases:', found_mail.body_html)
        self.assertIn('saas-database-1: database not found', found_mail.body_html)
        self.assertIn('saas-database-2: database not found', found_mail.body_html)

    @patch('requests.post')
    @patch('odoo.addons.openerp_enterprise.models.multidb_jobqueue.SaasAuth')
    @patch('odoo.addons.openerp_enterprise.models.auth_oauth2_token.AuthOauth2Token._get_trial_token')
    @patch('odoo.addons.openerp_enterprise.models.multidb_jobqueue.MAX_DBS_PER_REQUEST', new=1)
    def test_run_invite_users_by_small_batches(self, mock_get_trial_token, MockSaasAuth, mock_requests_post):
        mock_requests_post.return_value.json.return_value = {'result': {
            self.database1.db_uuid: {'result': True},
            self.database2.db_uuid: {'result': True},
        }}
        token1, token2 = sentinel.token1, sentinel.token2
        mock_get_trial_token.side_effect = [token1, token2]

        with self.mock_mail_gateway():
            self.job1.line_ids._run_invite_users(self.IPS_BY_SERVER_NAME['eu1.odoo.test'], 1337)

        url = 'http://eu1.odoo.test/saas_worker/multi/create_invitation/json'
        headers = {'X-Odoo-Port': '1337'}
        expected_data1 = {'params': {'databases': [
            {'db_uuid': self.database1.db_uuid, 'oauth_token': token1, 'user_login': 'admin@database1'}], 'invitees': []}}
        expected_data2 = {'params': {'databases': [
            {'db_uuid': self.database2.db_uuid, 'oauth_token': token2, 'user_login': 'admin@database2'}], 'invitees': []}}

        mock_requests_post.assert_has_calls([
            call(url, headers=headers, json=expected_data1, auth=MockSaasAuth(), timeout=15),
            call(url, headers=headers, json=expected_data2, auth=MockSaasAuth(), timeout=15),
        ], any_order=True)


class TestMultidbJobQueueRemoveUsers(TestMultidbJobQueue, HttpCase, MailCommon):

    def setUp(self):
        super().setUp()

        self.other_user1 = new_test_user(self.env, login='other_user1', groups='base.group_portal')
        self.other_user2 = new_test_user(self.env, login='other_user2', groups='base.group_portal')

        self.database1.write({
            'database_user_ids': [
                odoo.fields.Command.create({'user_id': self.other_user1.id, 'login': 'other_user1@database1'}),
                odoo.fields.Command.create({'user_id': self.other_user2.id, 'login': 'other_user2@database1'}),
            ],
        })

        self.database2.write({
            'database_user_ids': [
                odoo.fields.Command.create({'user_id': self.other_user1.id, 'login': 'other_user1@database1'}),
                odoo.fields.Command.create({'user_id': self.other_user2.id, 'login': 'other_user2@database1'}),
            ],
        })

        self.job1.write({
            'type': 'remove_users',
            'arguments': json.dumps({
                'removed_uids': [str(u.id) for u in [self.other_user1, self.other_user2]],
            }),
        })

    def test_controller(self):
        self.authenticate(self.user1.login, self.user1.login)

        previous_jobs = self.MultidbJob.search([])
        previous_cron_triggers = self.env['ir.cron.trigger'].search([])

        data = {'params': {
            'db_uuids': [self.database1.db_uuid, self.database2.db_uuid],
            'user_ids': [self.other_user1.id, self.other_user2.id],
        }}
        headers = {
            'Content-Type': 'application/json',
        }
        res = self.url_open('/odoo-enterprise/remove-users-multidb', data=json.dumps(data), headers=headers)
        res.raise_for_status()
        self.assertEqual(res.json()['result'], {'result': True})

        new_jobs = self.MultidbJob.search([]) - previous_jobs
        self.assertRecordValues(new_jobs, [{
            'user_id': self.user1.id,
            'impostor_id': False,
            'type': 'remove_users',
            'arguments': json.dumps({
                'removed_uids': [self.other_user1.id, self.other_user2.id],
            }),
            'state': 'todo',
        }])
        self.assertRecordValues(new_jobs.line_ids, [{
            'job_id': new_jobs.id,
            'database_id': self.database1.id,
            'state': 'todo',
        }, {
            'job_id': new_jobs.id,
            'database_id': self.database2.id,
            'state': 'todo',
        }])

        # The cron job has been triggered
        cron_triggers = self.env['ir.cron.trigger'].search([]) - previous_cron_triggers
        self.assertRecordValues(cron_triggers, [{
            'cron_id': self.env.ref('openerp_enterprise.cron_multidb_jobs').id,
            'call_at': DateBefore(odoo.fields.Datetime.now()),
        }])

    def test_controller_with_impostor(self):
        some_admin = new_test_user(self.env, login='some_admin', groups='base.group_system')
        self.authenticate(self.user1.login, self.user1.login)

        # Patch the session with impostor info
        self.session.impostor = some_admin.id
        self.session.impostor_info = some_admin.get_support_info()
        odoo.http.root.session_store.save(self.session)

        previous_jobs = self.MultidbJob.search([])

        data = {'params': {
            'db_uuids': [self.database1.db_uuid, self.database2.db_uuid],
            'user_ids': [self.other_user1.id, self.other_user2.id],
        }}
        headers = {
            'Content-Type': 'application/json',
        }
        res = self.url_open('/odoo-enterprise/remove-users-multidb', data=json.dumps(data), headers=headers)
        res.raise_for_status()
        self.assertEqual(res.json()['result'], {'result': True})

        new_jobs = self.MultidbJob.search([]) - previous_jobs
        self.assertRecordValues(new_jobs, [{
            'user_id': self.user1.id,
            'impostor_id': some_admin.id,
        }])

    @patch('requests.post')
    @patch('odoo.addons.openerp_enterprise.models.multidb_jobqueue.SaasAuth')
    @patch('odoo.addons.openerp_enterprise.models.auth_oauth2_token.AuthOauth2Token._get_trial_token')
    def test_run_remove_users(self, mock_get_trial_token, MockSaasAuth, mock_requests_post):
        mock_requests_post.return_value.json.return_value = {'result': {
            self.database1.db_uuid: {'result': True},
            self.database2.db_uuid: {'result': True},
        }}
        token1, token2 = sentinel.token1, sentinel.token2
        mock_get_trial_token.side_effect = [token1, token2]

        with self.mock_mail_gateway():
            self.job1.line_ids._run_remove_users(self.IPS_BY_SERVER_NAME['eu1.odoo.test'], 1337)

        session_info = {
            'impostor': False,
            'login': 'saas-user1@tld.net',
            'uid': self.user1.id,
        }
        mock_get_trial_token.assert_has_calls([
            call(self.user1.id, self.database1.db_uuid, session_info=session_info),
            call(self.user1.id, self.database2.db_uuid, session_info=session_info),
        ])

        expected_data = {'params': {
            'databases': [
                {'db_uuid': self.database1.db_uuid, 'oauth_token': token1, 'user_login': 'admin@database1'},
                {'db_uuid': self.database2.db_uuid, 'oauth_token': token2, 'user_login': 'admin@database2'},
            ],
            'user_oauth_ids': [str(self.other_user1.id), str(self.other_user2.id)],
        }}
        mock_requests_post.assert_called_once_with('http://eu1.odoo.test/saas_worker/multi/remove_users/json',
                                                   headers={'X-Odoo-Port': '1337'}, json=expected_data,
                                                   auth=MockSaasAuth(), timeout=30)
        self.assertNotSentEmail()
        self.assertEqual(self.job1.line_ids.mapped('state'), ['done', 'done'])
        self.assertEqual(self.job1.line_ids.mapped('result'), ['true', 'true'])

    @patch('requests.post')
    @patch('odoo.addons.openerp_enterprise.models.multidb_jobqueue.SaasAuth')
    @patch('odoo.addons.openerp_enterprise.models.auth_oauth2_token.AuthOauth2Token._get_trial_token')
    def test_run_remove_users_fails_db(self, mock_get_trial_token, MockSaasAuth, mock_requests_post):
        mock_requests_post.return_value.json.return_value = {'result': {
            self.database1.db_uuid: {'result': False, 'some': 'explanation'},
            self.database2.db_uuid: {'result': True},
        }}

        with self.mock_mail_gateway():
            self.job1.line_ids._run_remove_users(self.IPS_BY_SERVER_NAME['eu1.odoo.test'], 1337)

        self.assertEqual(self.job1.line_ids.mapped('state'), ['error', 'done'])
        self.assertEqual(self.job1.line_ids.mapped('result'), [
            '{"result": false, "some": "explanation"}',
            'true',
        ])

    @patch('requests.post')
    @patch('odoo.addons.openerp_enterprise.models.multidb_jobqueue.SaasAuth')
    @patch('odoo.addons.openerp_enterprise.models.auth_oauth2_token.AuthOauth2Token._get_trial_token')
    def test_run_remove_users_fails_db_not_found(self, mock_get_trial_token, MockSaasAuth, mock_requests_post):
        mock_requests_post.return_value.json.return_value = {'result': {
            self.database1.db_uuid: {'result': False, 'error_code': 'db_not_found'},
            self.database2.db_uuid: {'result': True},
        }}

        with self.mock_mail_gateway():
            self.job1.line_ids._run_remove_users(self.IPS_BY_SERVER_NAME['eu1.odoo.test'], 1337)

        self.assertEqual(self.job1.line_ids.mapped('state'), ['db_not_found', 'done'])
        self.assertEqual(self.job1.line_ids.mapped('result'), [
            '{"result": false, "error_code": "db_not_found"}',
            'true',
        ])

    @patch('requests.post')
    @patch('odoo.addons.openerp_enterprise.models.multidb_jobqueue.SaasAuth')
    @patch('odoo.addons.openerp_enterprise.models.auth_oauth2_token.AuthOauth2Token._get_trial_token')
    def test_run_remove_users_general_failure(self, mock_get_trial_token, MockSaasAuth, mock_requests_post):
        mock_requests_post.return_value.json.return_value = {'result': False, 'any': 'explanation'}

        with self.mock_mail_gateway():
            self.job1.line_ids._run_remove_users(self.IPS_BY_SERVER_NAME['eu1.odoo.test'], 1337)

        self.assertEqual(self.job1.line_ids.mapped('state'), ['error', 'error'])
        self.assertEqual(self.job1.line_ids.mapped('result'), [json.dumps({'result': False, 'any': 'explanation'})] * 2)

    def test_finalize_remove_users(self):
        self.job1.line_ids.write({
            'result': 'true',
            'state': 'done',
        })
        self.assertFalse(self.job1.result)
        self.assertEqual(self.job1.state, 'todo')

        with self.mock_mail_gateway():
            self.job1.line_ids._finalize_remove_users()

        self.assertTrue(self.job1.result)
        self.assertEqual(self.job1.state, 'done')

        # Check that a confirmation email is sent to the initiator user
        found_mail = self._find_mail_mail_wemail('"Saas User 1" <user1@tld.net>', 'outgoing', author=self.env.user.partner_id)
        self.assertIn('Access to several databases has been removed.', found_mail.body_html)

    @patch('requests.post')
    @patch('odoo.addons.openerp_enterprise.models.multidb_jobqueue.SaasAuth')
    @patch('odoo.addons.openerp_enterprise.models.auth_oauth2_token.AuthOauth2Token._get_trial_token')
    def test_run_remove_users_db_not_found(self, mock_get_trial_token, MockSaasAuth, mock_requests_post):
        """When a database is not found, it is ignored and _finalize is called"""
        mock_requests_post.return_value.json.return_value = {'result': {
            self.database1.db_uuid: {'result': False, 'error_code': 'db_not_found'},
            self.database2.db_uuid: {'result': True},
        }}

        with self.mock_mail_gateway():
            self.job1.line_ids._run_remove_users(self.IPS_BY_SERVER_NAME['eu1.odoo.test'], 1337)

        self.assertEqual(self.job1.state, 'todo')
        self.assertEqual(self.job1.line_ids.mapped('state'), ['db_not_found', 'done'])
        self.assertNotSentEmail()

        with self.mock_mail_gateway():
            self.job1.line_ids._finalize_remove_users()

        self.assertEqual(self.job1.state, 'done')
        self.assertEqual(self.job1.line_ids.mapped('state'), ['db_not_found', 'done'])

        found_mail = self._find_mail_mail_wemail('"Saas User 1" <user1@tld.net>', 'outgoing', author=self.env.user.partner_id)
        self.assertIn('Access to several databases has been removed.', found_mail.body_html)
        self.assertIn('However, the action could not be processed on these databases:', found_mail.body_html)
        self.assertIn('saas-database-1: database not found', found_mail.body_html)

    @patch('requests.post')
    @patch('odoo.addons.openerp_enterprise.models.multidb_jobqueue.SaasAuth')
    @patch('odoo.addons.openerp_enterprise.models.auth_oauth2_token.AuthOauth2Token._get_trial_token')
    def test_run_remove_users_all_dbs_not_found(self, mock_get_trial_token, MockSaasAuth, mock_requests_post):
        """When a database is not found, it is ignored and _finalize is called"""
        mock_requests_post.return_value.json.return_value = {'result': {
            self.database1.db_uuid: {'result': False, 'error_code': 'db_not_found'},
            self.database2.db_uuid: {'result': False, 'error_code': 'db_not_found'},
        }}

        with self.mock_mail_gateway():
            self.job1.line_ids._run_remove_users(self.IPS_BY_SERVER_NAME['eu1.odoo.test'], 1337)

        self.assertEqual(self.job1.state, 'todo')
        self.assertEqual(self.job1.line_ids.mapped('state'), ['db_not_found', 'db_not_found'])
        self.assertNotSentEmail()

        with self.mock_mail_gateway():
            self.job1.line_ids._finalize_remove_users()

        self.assertEqual(self.job1.state, 'done')
        self.assertEqual(self.job1.line_ids.mapped('state'), ['db_not_found', 'db_not_found'])

        found_mail = self._find_mail_mail_wemail('"Saas User 1" <user1@tld.net>', 'outgoing', author=self.env.user.partner_id)
        self.assertNotIn('Access to several databases has been removed:', found_mail.body_html)
        self.assertIn('However, the action could not be processed on these databases:', found_mail.body_html)
        self.assertIn('saas-database-1: database not found', found_mail.body_html)
        self.assertIn('saas-database-2: database not found', found_mail.body_html)
